@@ -24,15 +24,17 @@
 #include <string.h>
 #include <pcre.h>
 #include <sys/types.h>
+#include <sys/stat.h>
 #include <sys/wait.h>
 #include <fcntl.h>
 #include <stdarg.h>
 #include <dirent.h>
 #include <fnmatch.h>
 #include "dvdspanky.h"
+#include "vstr.h"
 #include "config.h"
 
-static char gp_title[] = "dvdspanky 0.10.1 - (C) 2007 Jeffrey Grembecki\n\n";
+static char gp_title[] = "dvdspanky 1.0.0 - (C) 2007 Jeffrey Grembecki\n\n";
 static char gp_licence[] =
 "This program is free software; you can redistribute it and/or modify\n"
 "it under the terms of the GNU General Public License as published by\n"
@@ -50,17 +52,21 @@ static char gp_licence[] =
 "MA 02110-1301, USA.\n";
 
 static char gp_help[] =
-"%s -i file [ -o file ] [-12bBFhILnOmNpTv] [-s size] [-r rate] [-R rate] [-c channels] [-V pixels] [-H pixels] [-a mode] [-f filters]\n"
+"%s -i file [OPTIONS]\n"
 "\n"
 "following; [ ] = default, { } = requirements, * = experimental\n"
 "\n"
+"-0, --avpass           auto volume pass only [off]\n"
 "-1, --firstpass        1st pass only [off]\n"
 "-2, --secondpass       2nd pass only including multiplex [off]\n"
 "-a, --aspect=MODE      aspect ratio 2=4:3, 3=16:9 [auto] {2,3}\n"
+"-A, --autovolume       automatically adjust volume [off]\n"
 "-b, --border           automatic black border [off]\n"
 "-B, --clip             clips edges to fit aspect [off]\n"
 "-c, --channels=CHANS   number of audio channels [auto] {1 - 5}\n"
 "-f, --filters=FILTERS  specify aditional transcode filters [none]\n"
+"-g, --gain=GAIN        multiply volume by GAIN [off/1.0]\n"
+"-G, --gamma=GAMMA      set gamma, < 1.0 is brighter, > is darker [1.0]\n"
 "-F, --frameadjust      interlace video for framerate conversion [off] *\n"
 "-h, --help             this help text\n"
 "-H, --hcrop=PIXELS     as above [off]\n"
@@ -72,7 +78,7 @@ static char gp_help[] =
 "-n, --ntsc             set ntsc [auto]\n"
 "-N, --ntscfix          fix variable framerate ntsc video [off] *\n"
 "-o, --output=FILE      output filename prefix (w/o extension)\n"
-"-O, --normalise        normalise audio volume [off]\n"
+"-O, --normalise        dynamically normalise audio volume [off]\n"
 "-r, --vrate=RATE       video bitrate [auto] {600 - 7500}\n"
 "-p, --pal              set pal [auto]\n"
 "-P, --preview          previews 5 frames using feh then exits [off]\n"
@@ -85,14 +91,18 @@ static char gp_help[] =
 
 static struct option gp_longopts[] =
 {
+	{"avpass", 0, 0, '0'},
 	{"firstpass", 0, 0, '1'},
 	{"secondpass", 0, 0, '2'},
 	{"aspect", 1, 0, 'a'},
+	{"autovolume", 0, 0, 'A'},
 	{"border", 0, 0, 'b'},
 	{"clip", 0, 0, 'B'},
 	{"channels", 1, 0, 'c'},
 	{"filters", 1, 0, 'f'},
 	{"frameadjust", 0, 0, 'F'},
+	{"gain", 1, 0, 'g'},
+	{"gamma", 1, 0, 'G'},
 	{"hcrop", 1, 0, 'H'},
 	{"help", 0, 0, 'h'},
 	{"input", 1, 0, 'i'},
@@ -116,48 +126,101 @@ static struct option gp_longopts[] =
 	{"verbose", 0, 0, 'v'},
 };
 
-static char gp_shortopts[] = "12a:bBc:f:FhH:i:ILmnNo:OpPr:R:s:TvV:";
+static char gp_shortopts[] = "012a:AbBc:f:Fg:G:hH:i:ILmnNo:OpPr:R:s:TvV:";
 
-pcre *gp_idregex = NULL;
-pcre *gp_transregex = NULL;
-pcre *gp_tcvideoregex = NULL;
-pcre *gp_tcaudioregex = NULL;
+static pcre *gp_idregex = NULL;
+static pcre *gp_transregex = NULL;
+static pcre *gp_tcvideoregex = NULL;
+static pcre *gp_tcaudioregex = NULL;
+
+#define ML_ERROR 4
+#define ML_WARNING 3
+#define ML_INFO 2
+#define ML_VERBOSE 1
+#define ML_DEBUG 0
+
+static int g_msglevel = ML_INFO;
+
+static FILE *gp_logfile = NULL;
+
+static void print(int msglevel, const char *p_format, ...)
+{
+	static vstr *p_str = NULL;
+	va_list argp;
+
+	if(!p_str)
+	{
+		if(!(p_str = vs_new()))
+			return;
+	}
+
+	if(msglevel >= g_msglevel)
+	{
+		va_start(argp, p_format);
+		vs_vprint(p_str, p_format, argp);
+		va_end(argp);
+		printf("%s", p_str->s);
+	}
+
+	if(gp_logfile)
+	{
+		vs_print(p_str, " dvdspanky | ");
+		va_start(argp, p_format);
+		vs_vcat(p_str, p_format, argp);
+		va_end(argp);
+		fwrite(p_str->s, 1, p_str->len, gp_logfile);
+		fflush(gp_logfile);
+	}
+}
+
+static void pprint(int msglevel, const char *p_prefix, const char *p_format, ...)
+{
+	static vstr *p_str = NULL;
+	va_list argp;
+
+	if(!p_str)
+	{
+		if(!(p_str = vs_new()))
+			return;
+	}
+
+	vs_print(p_str, "%10s | ", p_prefix);
+	va_start(argp, p_format);
+	vs_vcat(p_str, p_format, argp);
+	va_end(argp);
+
+	if(msglevel >= g_msglevel)
+		printf("%s", p_str->s);
+
+	if(gp_logfile)
+	{
+		fwrite(p_str->s, 1, p_str->len, gp_logfile);
+		fflush(gp_logfile);
+	}
+}
 
 static void printerror(const char *p_format, ...)
 {
+	static vstr *p_str = NULL;
 	va_list argp;
 
-	fprintf(stderr, "[dvdspanky] error: ");
+	if(!p_str)
+	{
+		if(!(p_str = vs_new()))
+			return;
+	}
+
+	vs_print(p_str, "error: ");
 	va_start(argp, p_format);
-	vfprintf(stderr, p_format, argp);
+	vs_vcat(p_str, p_format, argp);
 	va_end(argp);
-	fprintf(stderr, "\n");
-}
+	vs_cat(p_str, "\n");
 
-static void printwarning(const char *p_format, ...)
-{
-	va_list argp;
-
-	fprintf(stderr, "[dvdspanky] warning: ");
-	va_start(argp, p_format);
-	vfprintf(stderr, p_format, argp);
-	va_end(argp);
-	fprintf(stderr, "\n");
-}
-
-static void printverbose(const char *p_format, ...)
-{
-	va_list argp;
-
-	fprintf(stderr, "[dvdspanky] verbose: ");
-	va_start(argp, p_format);
-	vfprintf(stderr, p_format, argp);
-	va_end(argp);
-	fprintf(stderr, "\n");
+	print(ML_ERROR, p_str->s);
 }
 
 /* termination cleanup */
-void cleanup()
+static void cleanup()
 {
 	if(gp_idregex)
 		pcre_free(gp_idregex);
@@ -166,152 +229,116 @@ void cleanup()
 	if(gp_tcvideoregex)
 		pcre_free(gp_tcvideoregex);
 	gp_idregex = gp_transregex = gp_tcvideoregex = NULL;
+
+	if(gp_logfile)
+		fclose(gp_logfile);
+
+	gp_logfile = NULL;
 }
 
-/* prepare encoding */
+/* encode / preview */
 int encode(video_t *p_in, video_t *p_out, int options, const char *p_filters)
 {
-	char p_cmdline[4096];
-	char p_print[1000];
-	char p_num[20];
-	int asr;
+	vstr *p_generic, *p_video, *p_videoextra, *p_audio, *p_final;
 
-	p_cmdline[0] = 0;
+	p_generic = vs_new();
+	p_video = vs_new();
+	p_videoextra = vs_new();
+	p_audio = vs_new();
+	p_final = vs_new();
 
-	/* handy safe cat macros */
-#define safecat(_s) strncat(p_cmdline, _s, sizeof(p_cmdline) - strlen(p_cmdline) - 1)
-#define safecati(_n) { snprintf(p_num, sizeof(p_num), "%d", _n); \
-	strncat(p_cmdline, p_num, sizeof(p_cmdline) - strlen(p_cmdline) - 1); }
-#define safecatf(_f) { snprintf(p_num, sizeof(p_num), "%f", _f); \
-	strncat(p_cmdline, p_num, sizeof(p_cmdline) - strlen(p_cmdline) - 1); }
+	if(!p_generic || !p_video || !p_videoextra || !p_audio || !p_final)
+	{
+		if(p_generic) vs_free(p_generic);
+		if(p_video) vs_free(p_video);
+		if(p_videoextra) vs_free(p_videoextra);
+		if(p_audio) vs_free(p_audio);
+		if(p_final) vs_free(p_final);
+		printerror("unable to allocate strings");
+		return 1;
+	}
 
-	/* prepare command line */
-	p_cmdline[0] = p_print[0] = p_num[0] = 0;
-	safecat("transcode");
+#define freestrings() { vs_free(p_generic); vs_free(p_video); vs_free(p_videoextra); \
+	vs_free(p_audio); vs_free(p_final); }
 
-	/* nice */
+	/* generic */
+	vs_cat(p_generic, "transcode -i %s", p_in->p_filename);
 	if(!(options & SPANK_NONICE))
-		safecat(" --nice 19");
-
-	/* infile */
-	safecat(" -i ");
-	safecat(p_in->p_filename);
+		vs_cat(p_generic, " --nice 19");
 
 	/* audio */
-	if(p_out->achans)
-	{
-		safecat(" -E 48000,");
-		safecati(p_out->abits);
-		safecat(",");
-		safecati(p_out->achans);
-	}
-	else
-	{
-		safecat(" -E 48000");
-	}
-	safecat(" -N 0x2000 -b ");
-	safecati(p_out->arate);
-
-	/* filters */
-	if(options & SPANK_DOUBLEFPS)
-		safecat(" -J doublefps");
-	if(options & SPANK_PPROCESS)
-		safecat(" -J pp=de");
-	if(options & SPANK_NORMALISE)
-		safecat(" -J normalize");
-	if(p_filters[0])
-	{
-		safecat(" -J ");
-		safecat(p_filters);
-	}
+	vs_cat(p_audio, " -E 48000,%d,%d -N 0x2000 -b %d",
+		(p_out->abits) ? p_out->abits : 16, p_out->achans, p_out->arate);
 
 	/* video */
-	if(p_out->aspect < 1.5555)
-		asr = 2;
-	else
-		asr = 3;
-	snprintf(p_print, sizeof(p_print) - 1,
-		" -w %d --export_fps %.3f --export_asr %d --import_asr %d",
-		p_out->vrate, p_out->fps, asr, asr);
-	safecat(p_print);
-	if(p_out->p_vcodec[0] == 'P')
-		safecat(" --encode_fields t");
-	else
-		safecat(" --encode_fields b");
-
-	/* alter fps */
+	{
+		int asr = (p_out->aspect < 1.5) ? 2 : 3;
+		vs_cat(p_video, " -w %d -f %.3f --export_fps %.3f --import_asr %d --export_asr %d"
+			" --encode_fields %c", 	p_out->vrate, p_in->fps, p_out->fps, asr, asr,
+			(p_out->p_vcodec[0] == 'P') ? 't' : 'b');
+	}
+	if(options & SPANK_DOUBLEFPS)
+		vs_cat(p_video, " -J doublefps");
 	if(options & SPANK_NEWFPS)
-	{
-		safecat(" -M 0 -J ivtc,32detect=force_mode=3,decimate -f ");
-		safecatf(p_in->fps);
-	}
-
-	/* ntsc fix */
+		vs_cat(p_video, " -M 0 -J ivtc,32detect=force_mode=3,decimate");
 	if(options & SPANK_FIXNTSC)
-	{
-		safecat(" -M 2 -J ivtc,32detect=force_mode=5,decimate --hard_fps");
-	}
-
-	/* croping / bordering */
-	if(p_out->vcrop || p_out->hcrop)
-	{
-		snprintf(p_print, sizeof(p_print), " -j %d,%d,%d,%d",
-			p_out->vcrop, p_out->hcrop, p_out->vcrop, p_out->hcrop);
-		safecat(p_print);
-	}
-
-	/* mplayer input if transcode doesn't understand the input */
+		vs_cat(p_video, " -M 2 -J ivtc,32detect=force_mode=5,decimate --hard_fps");
 	if(options & SPANK_MPLAYER)
 	{
 		unlink("stream.yuv");
-		snprintf(p_print, sizeof(p_print),
-			" -H 0 -x mplayer,mplayer -f %.3f -g %dx%d -n 0x1 -e %d,%d,%d",
-			p_in->fps, p_in->width, p_in->height, p_in->hz,
-			(p_in->abits) ? p_in->abits : 16, p_in->achans);
-		safecat(p_print);
+		vs_cat(p_video, " -H 0 -x mplayer,mplayer -g %dx%d -n 0x1 -e %d,%d,%d", p_in->width,
+			p_in->height, p_in->hz,	(p_in->abits) ? p_in->abits : 16, p_in->achans);
 	}
 
+	/* video extra */
+	if(p_out->vcrop || p_out->hcrop)
+		vs_cat(p_videoextra, " -j %d,%d,%d,%d",
+			p_out->vcrop, p_out->hcrop, p_out->vcrop, p_out->hcrop);
+	if(p_out->gamma != 1.0)
+		vs_cat(p_videoextra, " -G %f", p_out->gamma);
+	if(options & SPANK_PPROCESS)
+		vs_cat(p_videoextra, " -J pp=de");
+	if(options & SPANK_NORMALISE)
+		vs_cat(p_videoextra, " -J normalize=algo=2 -s 0.5");
+	if(p_filters[0])
+		vs_cat(p_videoextra, " -J %s", p_filters);
+
+	/* create ffmpeg.cfg */
+	system("echo \"[mpeg2video]\nvrc_minrate=0\nvrc_maxrate=7500\nvrc_buf_size=1792\n\""
+		" > ffmpeg.cfg");
+
 	/* preview settings */
 	if(options & SPANK_PREVIEW)
 	{
 		struct dirent **pp_dirent;
 		int count;
 
-		p_out->options = options & SPANK_VERBOSE;
-		printf("generating previews..\n");
-		if(forkitf(NULL, NULL, TRANSCODE_BIN,
-			"%s -y jpg,null -F 100 -Z %dx%d -c 300-301,600-601,900-901,1200-1201,1500-1501 -o pre",
-			p_cmdline, (int)((double)p_out->height * p_out->aspect), p_out->height,
-			p_out->p_filename))
+		print(ML_INFO, "generating previews..\n"); /* generate previews */
+		if(forkitf(NULL, NULL, TRANSCODE_BIN, "%s %s %s -y jpg,null -F 100 -Z %dx%d"
+			" -c 300-301,600-601,900-901,1200-1201,1500-1501 -o pre",
+			p_generic->s, p_video->s, p_videoextra->s,
+			(int)((double)p_out->height * p_out->aspect), p_out->height, p_out->p_filename))
+		{
+			printerror("failed generating previews");
+			freestrings();
 			return 1;
+		}
 
 		count = scandir(".", &pp_dirent, 0, 0);
-		if(count >= 0)
+		if(count >= 0) /* display then remove previews */
 		{
-			printf("displaying previews..\n");
-			char p_cmd[512];
-			int cmdlen, found;
+			int found = count;
 
-			found = count;
+			print(ML_INFO, "displaying previews..\n");
 
-			/* construct feh command line */
-			strcpy(p_cmd, "feh");
-			cmdlen = strlen(p_cmd);
+			vs_print(p_final, "feh");
 			while(count--)
 			{
 				if(!fnmatch("pre*.jpg", pp_dirent[count]->d_name, 0))
-				{
-					strncat(p_cmd, " ", sizeof(p_cmd) - cmdlen - 1);
-					cmdlen++;
-					strncat(p_cmd, pp_dirent[count]->d_name, sizeof(p_cmd) - cmdlen - 1);
-					cmdlen += strlen(pp_dirent[count]->d_name);
-				}
+					vs_cat(p_final, " %s", pp_dirent[count]->d_name);
 			}
+			forkit(NULL, NULL, FEH_BIN, p_final->s);
 
-			/* start feh */
-			forkit(NULL, NULL, FEH_BIN, p_cmd);
-
-			/* remove preview images */
 			count = found;
 			while(count--)
 			{
@@ -321,78 +348,201 @@ int encode(video_t *p_in, video_t *p_out, int options, const char *p_filters)
 
 			free(pp_dirent);
 		}
+		freestrings();
 		return 0;
 	}
 
-	/* create ffmpeg.cfg */
-	system("echo \"[mpeg2video]\nvrc_minrate=0\nvrc_maxrate=7500\nvrc_buf_size=1792\n\" > ffmpeg.cfg");
+	/* auto volume only pass */
+	if(options & SPANK_AVPASS)
+	{
+		print(ML_INFO, "scanning volume..\n");
+		vs_print(p_final, "%s-tmp.ac3", p_out->p_filename);
+		if(forkitf(cb_transcode, (void*)p_out, TRANSCODE_BIN,
+			"%s %s -y null,ac3 -m %s-tmp -J astat=%s.norm",
+			p_generic->s, p_audio->s, p_out->p_filename, p_out->p_filename))
+		{
+			printerror("error scanning volume");
+			unlink(p_final->s);
+			freestrings();
+			return 1;
+		}
+		unlink(p_final->s);
+		return 0;
+	}
 
 	/* execute 1st pass */
 	if(options & SPANK_FIRST && !(options & SPANK_MULTIPLEX))
 	{
-		printf("starting 1st VBR pass..\n");
-		p_out->options = (options & SPANK_VERBOSE) | SPANK_FIRST;
-		if(forkitf(cb_transcode, (void*)p_out, TRANSCODE_BIN,
-			"%s -Z %dx%d -y ffmpeg,null -F mpeg2video -R 1,%s.pass1",
-			p_cmdline, p_out->width, p_out->height, p_out->p_filename))
+		print(ML_INFO, "starting 1st VBR pass..\n");
+
+		vs_print(p_final, "%s %s %s", p_generic->s, p_video->s, p_audio->s);
+		if(options & SPANK_AVOLUME)
+			vs_cat(p_final,
+				" -y ffmpeg -F mpeg2video -R 1,%s.pass1 -m %s.tmp -J astat=%s.norm",
+				p_out->p_filename, p_out->p_filename, p_out->p_filename);
+		else
+			vs_cat(p_final, " -y ffmpeg,null -F mpeg2video -R 1,%s.pass1", p_out->p_filename);
+
+		if(forkit(cb_transcode, (void*)p_out, TRANSCODE_BIN, p_final->s))
+		{
+			vs_print(p_final, "%s.tmp", p_out->p_filename);
+			unlink(p_final->s);
+			printerror("transcode failed");
+			freestrings();
 			return 1;
-		printf("\n");
+		}
+		vs_print(p_final, "%s.tmp", p_out->p_filename);
+		unlink(p_final->s);
+
+		print(ML_INFO, "\n");
 	}
 
 	/* execute 2nd pass */
 	if(options & SPANK_SECOND && !(options & SPANK_MULTIPLEX))
 	{
-		printf("starting 2nd VBR pass..\n");
-		p_out->options = (options & SPANK_VERBOSE) | SPANK_SECOND;
-		if(forkitf(cb_transcode, (void*)p_out, TRANSCODE_BIN,
-			"%s -Z %dx%d -y ffmpeg -F mpeg2video -R 2,%s.pass1 -o %s -m %s.ac3",
-			p_cmdline, p_out->width, p_out->height, p_out->p_filename,
-			p_out->p_filename, p_out->p_filename))
+		/* volume auto adjust */
+		if(options & SPANK_AVOLUME)
+		{
+			FILE *p_file;
+			char p_buffer[320];
+
+			vs_print(p_final, "%s.norm", p_out->p_filename);
+			p_file = fopen(p_final->s, "r");
+			if(!p_file)
+			{
+				printerror("unable to open %s.norm for autoaudio adjustment", p_out->p_filename);
+				freestrings();
+				return 1;
+			}
+			if(!fgets(p_buffer, sizeof(p_buffer) - 1, p_file))
+			{
+				printerror("%s.norm contains no relevant data", p_out->p_filename);
+				fclose(p_file);
+				freestrings();
+				return 1;
+			}
+			p_out->volume = atof(p_buffer);
+			fclose(p_file);
+		}
+
+		/* check pass1 file exists */
+		{
+			struct stat sdata;
+			vs_print(p_final, "%s.pass1", p_out->p_filename);
+			if(stat(p_final->s, &sdata))
+			{
+				printerror("unable to open %s.pass1", p_out->p_filename);
+				freestrings();
+				return 1;
+			}
+		}
+
+		/* transcode */
+		print(ML_INFO, "starting 2nd VBR pass..\n");
+		vs_print(p_final,
+			"%s %s %s %s -Z %dx%d -y ffmpeg -F mpeg2video -R 2,%s.pass1 -o %s -m %s.ac3",
+			p_generic->s, p_audio->s, p_video->s, p_videoextra->s,
+			p_out->width, p_out->height, p_out->p_filename, p_out->p_filename, p_out->p_filename);
+		if(p_out->volume != 1.0)
+			vs_cat(p_final, " -s %f", p_out->volume);
+		if(forkitf(cb_transcode, (void*)p_out, TRANSCODE_BIN, p_final->s))
+		{
+			printerror("transcode failed");
+			freestrings();
 			return 1;
-		printf("\n");
+		}
+		print(ML_INFO, "\n");
 	}
 
 	/* CBR encoding */
 	if(!(options & (SPANK_FIRST | SPANK_SECOND)) && !(options & SPANK_MULTIPLEX))
 	{
-		printf("starting CBR encode..\n");
-		p_out->options = options & SPANK_VERBOSE;
-		if(forkitf(cb_transcode, (void*)p_out, TRANSCODE_BIN,
-			"%s -Z %dx%d -y ffmpeg -F mpeg2video -o %s -m %s.ac3",
-			p_cmdline, p_out->width, p_out->height, p_out->p_filename, p_out->p_filename))
+		/* volume auto adjust */
+		if(options & SPANK_AVOLUME)
+		{
+			char p_buffer[320];
+			FILE *p_file;
+
+			print(ML_INFO, "scanning volume..\n");
+			forkitf(cb_transcode, (void*)p_out, TRANSCODE_BIN,
+				"%s %s -y null,ac3 -m %s-tmp -J astat=%s.norm",
+				p_generic->s, p_audio->s, p_out->p_filename, p_out->p_filename);
+
+			vs_print(p_final, "%s-tmp.ac3", p_out->p_filename);
+			unlink(p_final->s);
+
+			vs_print(p_final, "%s.norm", p_out->p_filename);
+			p_file = fopen(p_final->s, "r");
+			if(!p_file)
+			{
+				printerror("unable to open %s.norm for autoaudio adjustment", p_out->p_filename);
+				freestrings();
+				return 1;
+			}
+			if(!fgets(p_buffer, sizeof(p_buffer) - 1, p_file))
+			{
+				printerror("%s.norm contains no relevant data", p_out->p_filename);
+				fclose(p_file);
+				freestrings();
+				return 1;
+			}
+			p_out->volume = atof(p_buffer);
+			fclose(p_file);
+		}
+
+		/* execute */
+		print(ML_INFO, "starting CBR encode..\n");
+		vs_print(p_final, "%s %s %s %s -Z %dx%d -y ffmpeg -F mpeg2video -o %s -m %s.ac3",
+			p_generic->s, p_video->s, p_videoextra->s, p_audio->s,
+			p_out->width, p_out->height, p_out->p_filename, p_out->p_filename);
+		if(p_out->volume != 1.0)
+			vs_cat(p_final, " -s %f", p_out->volume);
+		if(forkit(cb_transcode, (void*)p_out, TRANSCODE_BIN, p_final->s))
+		{
+			printerror("transcode failed");
+			freestrings();
 			return 2;
-		printf("\n");
+		}
+		print(ML_INFO, "\n");
 	}
 
 	/* multiplex */
 	if((options & SPANK_SECOND) || !(options & (SPANK_FIRST | SPANK_SECOND)) ||
 		(options & SPANK_MULTIPLEX))
 	{
-		printf("multiplexing..\n");
+		print(ML_INFO, "multiplexing..\n");
 		if((options & SPANK_SECOND))
 		{
-			if(forkitf(NULL, NULL, MPLEX_BIN, "mplex -V -f 8 -o %s.mpg %s.m2v %s.ac3",
+			if(forkitf(NULL, "mplex |", MPLEX_BIN,
+				"mplex -V -f 8 -o %s.mpg %s.m2v %s.ac3",
 				p_out->p_filename, p_out->p_filename, p_out->p_filename))
-				return 3;
+				{
+					printerror("mplex failed");
+					freestrings();
+					return 3;
+				}
 		}
 		else
 		{
-			if(forkitf(NULL, NULL, MPLEX_BIN, "mplex -f 8 -o %s.mpg %s.m2v %s.ac3",
+			if(forkitf(NULL, "mplex |", MPLEX_BIN,
+				"mplex -f 8 -o %s.mpg %s.m2v %s.ac3",
 				p_out->p_filename, p_out->p_filename, p_out->p_filename))
+			{
+				printerror("mplex failed");
+				freestrings();
 				return 3;
+			}
 		}
 
-		snprintf(p_print, sizeof(p_print), "%s.m2v", p_out->p_filename);
-		unlink(p_print);
-		snprintf(p_print, sizeof(p_print), "%s.ac3", p_out->p_filename);
-		unlink(p_print);
+		vs_print(p_final, "%s.m2v");
+		unlink(p_final->s);
+		vs_print(p_final, "%s.ac3");
+		unlink(p_final->s);
 	}
 
-#undef safecat
-#undef safecati
-#undef safecatf
-
+	freestrings();
 	return 0;
+#undef freestrings
 }
 
 /* tcprobe identify callback */
@@ -414,6 +564,8 @@ int cb_tcprobe(const char *p_string, void *p_options)
 		return 0;
 	}
 
+	pprint(ML_VERBOSE, "tcprobe", "%s\n", p_string);
+
 	/* video regex */
 	length = strlen(p_string);
 	numstrings = pcre_exec(gp_tcvideoregex, NULL, p_string, length, 0, 0,
@@ -496,7 +648,6 @@ int cb_transcode(const char *p_string, void *p_options)
 {
 	char p_buffer[1024];
 	int numstrings, length, p_vector[20];
-	double videopos;
 	video_t *p_video;
 	int i;
 
@@ -516,19 +667,24 @@ int cb_transcode(const char *p_string, void *p_options)
 	if(numstrings > 0)
 	{
 		if(length >= sizeof(p_buffer))
+		{
+			printerror("string exceeds buffer");
 			return 1;
+		}
 		memcpy(p_buffer, p_string, length);
-		for(i = 1; i < 5; i++)
+		for(i = 1; i < 6; i++)
 			p_buffer[p_vector[i * 2 + 1]] = 0;
-		videopos = atof(&p_buffer[p_vector[4]]) * 3500.0 +
-			atof(&p_buffer[p_vector[6]]) * 60.0 +
-			atof(&p_buffer[p_vector[8]]);
-		fprintf(stderr,"  %7.2f fps %7.2f%% complete  \r", atof(&p_buffer[p_vector[2]]),
-			videopos * 100.0 / (double)p_video->length);
+		if(p_video->frames)
+			fprintf(stderr,"  %7.2f fps, %7.2f%% complete, frame %d  \r", atof(&p_buffer[p_vector[4]]),
+				atof(&p_buffer[p_vector[2]]) * 100.0 / (double)p_video->frames, atoi(&p_buffer[p_vector[2]]));
+		else
+			fprintf(stderr,"  %7.2f fps, frame %09d, %d:%02d:%02d  \r", atof(&p_buffer[p_vector[4]]),
+				atoi(&p_buffer[2]), atoi(&p_buffer[p_vector[4]]), atoi(&p_buffer[p_vector[6]]),
+				atoi(&p_buffer[p_vector[8]]));
 	}
-	else if(p_video->options & SPANK_VERBOSE)
+	else
 	{
-		printverbose("%s\n", p_buffer);
+		pprint(ML_VERBOSE, "transcode", "%s\n", p_string);
 	}
 
 	return 0;
@@ -550,6 +706,8 @@ int cb_mplayer(const char *p_string, void *p_options)
 	/* option is output to a video_t */
 	p_video = (video_t*)p_options;
 
+	pprint(ML_VERBOSE, "mplayer", "%s\n", p_string);
+
 	/* examine the string, return if nothing */
 	numstrings = pcre_exec(gp_idregex, NULL, p_string, strlen(p_string), 0, 0, p_vector, 20);
 	if(numstrings < 0)
@@ -560,42 +718,26 @@ int cb_mplayer(const char *p_string, void *p_options)
 	pcre_get_substring(p_string, p_vector, numstrings, 2, &p_value);
 
 	if(!strcmp(p_id, "VIDEO_FORMAT"))
-	{
 		strncpy(p_video->p_vcodec, p_value, sizeof(p_video->p_vcodec) - 1);
-	}
 	if(!strcmp(p_id, "VIDEO_BITRATE"))
-	{
 		p_video->vrate = atoi(p_value) / 1000;
-	}
 	if(!strcmp(p_id, "VIDEO_WIDTH"))
-	{
 		p_video->width = atoi(p_value);
-	}
 	if(!strcmp(p_id, "VIDEO_HEIGHT"))
 	{
 		p_video->height = atoi(p_value);
 		p_video->aspect = (float)p_video->width / (float)p_video->height;
 	}
 	if(!strcmp(p_id, "VIDEO_FPS"))
-	{
 		p_video->fps = atof(p_value);
-	}
 	if(!strcmp(p_id, "AUDIO_CODEC"))
-	{
 		strncpy(p_video->p_acodec, p_value, sizeof(p_video->p_acodec) - 1);
-	}
 	if(!strcmp(p_id, "AUDIO_BITRATE"))
-	{
 		p_video->arate = atoi(p_value) / 1000;
-	}
 	if(!strcmp(p_id, "AUDIO_RATE"))
-	{
 		p_video->hz = atoi(p_value);
-	}
 	if(!strcmp(p_id, "AUDIO_NCH"))
-	{
 		p_video->achans = atoi(p_value);
-	}
 	if(!strcmp(p_id, "LENGTH"))
 	{
 		p_video->length = atoi(p_value);
@@ -610,6 +752,20 @@ int cb_mplayer(const char *p_string, void *p_options)
 	return 0;
 }
 
+/* generic callback */
+int cb_default(const char *p_string, void *p_options)
+{
+
+	/* initialisation command */
+	if(!p_string)
+		return 0;
+
+	/* output le shiz */
+	pprint(ML_VERBOSE, "?", "%s %s\n", (char*)p_options, p_string);
+
+	return 0;
+}
+
 /* fork another program */
 int forkit(forkitcallback callback, void *p_options, const char *p_path, const char *p_cmd)
 {
@@ -624,10 +780,14 @@ int forkit(forkitcallback callback, void *p_options, const char *p_path, const c
 	cmdlength = strlen(p_cmd);
 	p_cmdline = malloc(cmdlength + 1);
 
+	if(!callback)
+		callback = cb_default;
+
 	/* create parameter index */
 	strncpy(p_cmdline, p_cmd, cmdlength + 1);
 	pos = start = 0;
-	/* fprintf(stderr, "[debug] fork %s, %s\n", p_path,p_cmd); */
+	print(ML_VERBOSE, "fork %s, %s\n", p_path,p_cmd);
+
 	for(i = 0; i < sizeof(pp_cmdopts) / sizeof(char**); i++)
 	{
 		while(p_cmdline[pos] != ' ' && p_cmdline[pos] != 0)
@@ -648,7 +808,10 @@ int forkit(forkitcallback callback, void *p_options, const char *p_path, const c
 
 	/* open comminication pipe */
 	if(pipe(p_pipe) == -1)
+	{
+		printerror("open pipe failed");
 		return 1;
+	}
 
 	/* fork */
 	pid = fork();
@@ -658,6 +821,7 @@ int forkit(forkitcallback callback, void *p_options, const char *p_path, const c
 			/* error */
 			close(p_pipe[0]);
 			close(p_pipe[1]);
+			printerror("fork %s failed", p_path);
 			retval = -2;
 			break;
 
@@ -667,7 +831,7 @@ int forkit(forkitcallback callback, void *p_options, const char *p_path, const c
 			dup2(p_pipe[1], fileno(stdout));
 			dup2(p_pipe[1], fileno(stderr));
 			execv(p_path, pp_cmdopts);
-			fprintf(stderr, "[dvdspanky] fork %s,%s failed\n", p_path,p_cmd);
+			printerror("fork %s, %s failed", p_path, p_cmd);
 			free(p_cmdline);
 			exit(1);
 
@@ -717,7 +881,7 @@ int forkit(forkitcallback callback, void *p_options, const char *p_path, const c
 			{
 				if(WIFEXITED(status))
 					retval = WEXITSTATUS(status);
-				else
+				else if(WIFSIGNALED(status))
 					retval = -1;
 			}
 
@@ -758,35 +922,36 @@ int main(int argc, char **argv)
 	/* clear video data */
 	memset(&invid, '\0', sizeof(video_t));
 	memset(&outvid, '\0', sizeof(video_t));
+	outvid.volume = outvid.gamma = 1.0;
 
 	atexit(cleanup);
 
 	/* show title */
-	printf(gp_title);
+	print(ML_INFO, gp_title);
 
 	/* compile regex */
 	if(!(gp_transregex = pcre_compile(
-		"^encoding\\sframes\\s.*?,\\s*(\\d+[.]\\d+)\\sfps.*?EMT:\\s*(\\d+):(\\d+):(\\d+),.*",
+		"^encoding\\sframes\\s\\[\\d+\\-(\\d+)\\],\\s*(\\d+[.]\\d+)\\sfps.*?EMT:\\s*(\\d+):(\\d+):(\\d+),.*",
 		0, &p_error, &erroffset, NULL)))
 	{
-		fprintf(stderr, "error: gp_transregex compile failed - %s\n", p_error);
+		printerror("gp_transregex compile failed - %s", p_error);
 		exit(1);
 	}
 	if(!(gp_idregex = pcre_compile("^ID_([^=]+)=(.*)$", 0, &p_error, &erroffset, NULL)))
 	{
-		fprintf(stderr, "error: gp_idregex compile failed - %s\n", p_error);
+		printerror("gp_idregex compile failed - %s", p_error);
 		exit(1);
 	}
 	if(!(gp_tcvideoregex = pcre_compile("^\\[.*?\\]\\sV:\\s([^ ]*)\\sfps,\\scodec=([^,]*),"
 		"\\sframes=([^,]*),\\swidth=([^,]*),\\sheight=(.*)$", 0, &p_error, &erroffset, NULL)))
 	{
-		fprintf(stderr, "error: gp_tcvideoregex compile failed - %s\n", p_error);
+		printerror("error: gp_tcvideoregex compile failed - %s", p_error);
 		exit(1);
 	}
 	if(!(gp_tcaudioregex = pcre_compile("^\\[.*?\\]\\sA:\\s([^ ]*)\\sHz,\\sformat=[^,]*,"
 		"\\sbits=([^,]*),\\schannels=([^,]*),\\sbitrate=(.*),$", 0, &p_error, &erroffset, NULL)))
 	{
-		fprintf(stderr, "error: gp_tcaudioregex compile failed - %s\n", p_error);
+		printerror("error: gp_tcaudioregex compile failed - %s", p_error);
 		exit(1);
 	}
 
@@ -795,10 +960,15 @@ int main(int argc, char **argv)
 	{
 		switch(opt)
 		{
+			case '0':
+				options |= SPANK_AVPASS;
+				break;
 			case '1':
+				options &= 0xffff ^ SPANK_SECOND;
 				options |= SPANK_FIRST;
 				break;
 			case '2':
+				options &= 0xffff ^ SPANK_FIRST;
 				options |= SPANK_SECOND;
 				break;
 			case 'a':
@@ -809,6 +979,9 @@ int main(int argc, char **argv)
 					exit(1);
 				}
 				break;
+			case 'A':
+				options |= SPANK_AVOLUME;
+				break;
 			case 'b':
 				calcborder = 1;
 				break;
@@ -829,6 +1002,12 @@ int main(int argc, char **argv)
 			case 'F':
 				options |= SPANK_NEWFPS;
 				break;
+			case 'g':
+				outvid.volume = atof(optarg);
+				break;
+			case 'G':
+				outvid.gamma = atof(optarg);
+				break;
 			case 'h':
 				printf(gp_help, argv[0]);
 				exit(0);
@@ -893,7 +1072,7 @@ int main(int argc, char **argv)
 				options |= SPANK_PPROCESS;
 				break;
 			case 'v':
-				options |= SPANK_VERBOSE;
+				g_msglevel = ML_VERBOSE;
 				break;
 			case 'V':
 				vcrop = atoi(optarg);
@@ -908,13 +1087,22 @@ int main(int argc, char **argv)
 		}
 	}
 
-	/* test input filename */
+	/* input file */
 	if(!invid.p_filename[0])
 	{
-		/* TODO: check for existance of file */
 		printerror("no input file");
 		exit(1);
 	}
+	{
+		struct stat sdata;
+		if(stat(invid.p_filename, &sdata))
+		{
+			printerror("cannot open input file %s", invid.p_filename);
+			exit(1);
+		}
+		invid.bytes = sdata.st_size;
+	}
+
 
 	/* create output filename based on input filename if none given */
 	if(!outvid.p_filename[0])
@@ -945,6 +1133,25 @@ int main(int argc, char **argv)
 		pcre_free(p_regex);
 	}
 
+	/* start log */
+	{
+		vstr *p_str;
+
+		if(!(p_str = vs_new()))
+		{
+			printerror("unable to create string");
+			exit(1);
+		}
+		vs_print(p_str, "%s.log", outvid.p_filename);
+		if(!(gp_logfile = fopen(p_str->s, "w")))
+		{
+			printerror("unable creating log file");
+			vs_free(p_str);
+			exit(1);
+		}
+		vs_free(p_str);
+	}
+
 	/* get source video details */
 	forkitf(cb_mplayer, &invid, MPLAYER_BIN,
 		"mplayer %s -identify -frames 0 -ao null -vo null", invid.p_filename);
@@ -970,6 +1177,13 @@ int main(int argc, char **argv)
 	outvid.abits = 16;
 	outvid.hz = 48000;
 
+	/* approximate missing data */
+	if(!invid.frames && invid.length)
+		invid.frames = (int)((double)invid.length * invid.fps);
+	if(!invid.vrate && invid.frames)
+		invid.vrate = (double)invid.bytes * 8.4 / ((double)invid.frames / invid.fps) / 1000.0 -
+			invid.arate;
+
 	/* calculate video bitrate for final size allowing 5% overhead */
 	if(calcsize)
 	{
@@ -1119,42 +1333,32 @@ int main(int argc, char **argv)
 	outvid.vcrop = vcrop;
 	outvid.hcrop = hcrop;
 
+	outvid.bytes = (vrate + vrate / 20 + arate + arate / 20) / 8 * invid.length * 1000;
+
 	/* print source and dest video encoding info */
-	printf(
-		" source | %s\n"
-		"  video | %s, %dx%d, %d kbps, %.3f fps\n"
-		"        | %.4f aspect, %d frames, %d:%02d:%02d\n"
-		"  audio | %s, %d kbps, %d Hz, %d bit, %d channel\n",
-		invid.p_filename,
-		invid.p_vcodec, invid.width, invid.height, invid.vrate, invid.fps,
-		invid.aspect, invid.frames,	invid.hours, invid.minutes, invid.seconds,
-		invid.p_acodec, invid.arate, invid.hz, invid.abits, invid.achans
-		);
+	print(ML_INFO, " source | %s\n", invid.p_filename);
+	print(ML_INFO, "  video | %s, %dx%d, %d kbps, %.3f fps\n",
+		invid.p_vcodec, invid.width, invid.height, invid.vrate, invid.fps);
+	print(ML_INFO, "        | %.4f aspect, %d frames, %d:%02d:%02d\n",
+		invid.aspect, invid.frames,	invid.hours, invid.minutes, invid.seconds);
+	print(ML_INFO, "  audio | %s, %d kbps, %d Hz, %d bit, %d channel\n",
+		invid.p_acodec, invid.arate, invid.hz, invid.abits, invid.achans);
+	print(ML_INFO, "   file | %d KB\n", invid.bytes / 1000);
 	if(options & SPANK_MPLAYER)
-		printf(
-		"   info | using mplayer decoder\n"
-		);
-	printf(
-		"\n"
-		);
+		print(ML_INFO, "   info | using mplayer decoder\n");
+	print(ML_INFO, "\n");
 
-
-	printf(
-		"   dest | %s.mpg\n"
-		"  video | %s, %dx%d, %d kbps, %.3f fps\n"
-		"        | %.4f aspect, %d frames, %d:%02d:%02d\n"
-		"        | %d:%d crop, %s\n"
-		"  audio | %s, %d kbps, %d Hz, %d bit, %d channel\n"
-		"   file | ~%d KB\n\n",
-		outvid.p_filename,
-		outvid.p_vcodec, outvid.width, outvid.height, outvid.vrate, outvid.fps,
-		outvid.aspect, outvid.frames, outvid.hours, outvid.minutes, outvid.seconds,
-		outvid.vcrop, outvid.hcrop, (options & SPANK_MULTI) ? "VBR" : "CBR",
-		outvid.p_acodec, outvid.arate, outvid.hz, outvid.abits, outvid.achans,
-		(vrate + vrate / 20 + arate + arate / 20) / 8 * invid.length
-		);
+	print(ML_INFO, "   dest | %s.mpg\n", outvid.p_filename);
+	print(ML_INFO, "  video | %s, %dx%d, %d kbps, %.3f fps\n",
+		outvid.p_vcodec, outvid.width, outvid.height, outvid.vrate, outvid.fps);
+	print(ML_INFO, "        | %.4f aspect, %d frames, %d:%02d:%02d\n",
+		outvid.aspect, outvid.frames, outvid.hours, outvid.minutes, outvid.seconds);
+	print(ML_INFO, "        | %d:%d crop, %s\n",
+		outvid.vcrop, outvid.hcrop, (options & SPANK_MULTI) ? "VBR" : "CBR");
+	print(ML_INFO, "  audio | %s, %d kbps, %d Hz, %d bit, %d channel\n",
+		outvid.p_acodec, outvid.arate, outvid.hz, outvid.abits, outvid.achans);
+	print(ML_INFO, "   file | ~%d KB\n", outvid.bytes / 1000);
+	print(ML_INFO, "\n");
 
 	exit(encode(&invid, &outvid, options, p_filters));
-
-	return 0;
 }
